
@file:OptIn(ExperimentalMaterial3Api::class)

package com.example.app.screens

import android.app.Application
import android.util.Log
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.widthIn
import androidx.compose.foundation.layout.wrapContentSize
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.MoreVert
import androidx.compose.material3.DropdownMenu
import androidx.compose.material3.DropdownMenuItem
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.material3.TopAppBar
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import androidx.lifecycle.viewmodel.compose.viewModel
import com.arcgismaps.Color
import com.arcgismaps.geometry.Envelope
import com.arcgismaps.geometry.Point
import com.arcgismaps.geometry.SpatialReference
import com.arcgismaps.mapping.ArcGISMap
import com.arcgismaps.mapping.BasemapStyle
import com.arcgismaps.mapping.GeoElement
import com.arcgismaps.mapping.Viewpoint
import com.arcgismaps.mapping.symbology.SimpleLineSymbol
import com.arcgismaps.mapping.symbology.SimpleLineSymbolStyle
import com.arcgismaps.mapping.symbology.SimpleMarkerSymbol
import com.arcgismaps.mapping.symbology.SimpleMarkerSymbolStyle
import com.arcgismaps.mapping.view.Graphic
import com.arcgismaps.mapping.view.GraphicsOverlay
import com.arcgismaps.mapping.view.SingleTapConfirmedEvent
import com.arcgismaps.tasks.geocode.GeocodeParameters
import com.arcgismaps.tasks.geocode.LocatorTask
import com.arcgismaps.toolkit.geoviewcompose.MapView
import com.arcgismaps.toolkit.geoviewcompose.MapViewProxy
import com.example.app.R
import kotlinx.coroutines.Job
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.asStateFlow
import kotlinx.coroutines.launch

class MapViewModel() : ViewModel() {

    // Create a map using the basemap style ArcGISTopographic.
    val map = ArcGISMap(BasemapStyle.ArcGISTopographic).apply {
        initialViewpoint = Viewpoint(
            center = Point(
                x = -118.477324,
                y = 33.999390,
                spatialReference = SpatialReference.wgs84()
            ),
            scale = 72_000.0
        )
    }
    val mapViewProxy = MapViewProxy()

    val graphicsOverlay = GraphicsOverlay()
    // The current extent visible in the map view.
    lateinit var geoViewExtent: Envelope

    // The location on the map that the user tapped to obtain further information. (A graphic is a type of GeoElement).
    private val _tapLocation = MutableStateFlow<Point?>(null)
    val tapLocation: StateFlow<Point?> = _tapLocation.asStateFlow()

    // The geoelement closest to the location that the user tapped.
    private val _selectedGeoElement = MutableStateFlow<GeoElement?>(null)
    val selectedGeoElement: StateFlow<GeoElement?> = _selectedGeoElement.asStateFlow()

    private var currentIdentifyJob: Job? = null

    val locator = LocatorTask(uri = "https://geocode-api.arcgis.com/arcgis/rest/services/World/GeocodeServer")

    enum class Category(val label: String, val color: Color) {
        // CoffeeShop color is brown.
        CoffeeShop(label = "Coffee shop", color = Color.fromRgba(r =150, g = 75, b= 0, a = 255)),
        // GasStation color is orange.
        GasStation(label = "Gas station", color = Color.fromRgba(r = 255, g = 165, b = 0, a =255)),
        // Food color is purple.
        Food(label = "Food", color = Color.fromRgba(r = 160, g = 32, b = 240, a = 255)),
        // Hotel color is blue.
        Hotel(label = "Hotel", color = Color.fromRgba(r = 0, g = 0, b = 255, a = 255)),
        ParksOutdoors(label = "Parks and Outdoors", Color.green)
    }

    suspend fun findPlaces(category: Category) {

        graphicsOverlay.graphics.clear()

        val geocodeParameters = GeocodeParameters().apply {
            searchArea = geoViewExtent
            resultAttributeNames.addAll(listOf("Place_addr", "PlaceName"))
        }

        val geocodeResultsList = locator.geocode(
            searchText = category.label,
            parameters = geocodeParameters
        ).getOrElse { error ->
            return logError(error)
        }

        if (geocodeResultsList.isNotEmpty()) {
            val placeSymbol = SimpleMarkerSymbol(
                style = SimpleMarkerSymbolStyle.Circle,
                color = category.color,
                size = 10f
            ).apply {
                outline = SimpleLineSymbol(
                    style = SimpleLineSymbolStyle.Solid,
                    color = Color.white,
                    width = 2f
                )
            }
            val graphics = geocodeResultsList.map { geocodeResult ->
                Graphic(
                    geometry = geocodeResult.displayLocation,
                    attributes = geocodeResult.attributes,
                    symbol = placeSymbol
                )
            }
            graphicsOverlay.graphics.addAll(graphics)
        }

    }

    /**
     * Identifies the tapped screen coordinate in the provided [singleTapConfirmedEvent]. The
     * identified geoelement is set to [_selectedGeoElement].
     *
     * @since 200.5.0
     */
    fun identify(singleTapConfirmedEvent: SingleTapConfirmedEvent) {
        currentIdentifyJob?.cancel()
        currentIdentifyJob = viewModelScope.launch {
            val result = mapViewProxy.identify(
                graphicsOverlay = graphicsOverlay,
                screenCoordinate = singleTapConfirmedEvent.screenCoordinate,
                tolerance = 2.dp
            )
            result.onSuccess { identifyGraphicsOverlayResult ->
                _selectedGeoElement.value = identifyGraphicsOverlayResult.geoElements.firstOrNull()
            }.onFailure { error ->
                logError(error)
            }
        }
    }

    fun clearSelectedGeoElement() {
        _selectedGeoElement.value = null
    }

    private fun logError(error: Throwable) {
        Log.e(this.javaClass.simpleName, error.message.toString(), error.cause)
    }

}

@Composable
fun MainScreen() {

    var isDropdownExpanded by remember { mutableStateOf(false) }

    // Create a ViewModel to handle MapView interactions.
    val mapViewModel: MapViewModel = viewModel()
    val selectedGeoElement = mapViewModel.selectedGeoElement.collectAsState().value

    Scaffold(
        topBar = {
            TopAppBar(
                title = { Text(text = stringResource(id = R.string.app_name)) },

                actions = {

                    IconButton(onClick = { isDropdownExpanded = !isDropdownExpanded }) {
                        Icon(
                            imageVector = Icons.Default.MoreVert,
                            contentDescription = "More options"
                        )
                    }

                    DropdownMenu(
                        expanded = isDropdownExpanded,
                        onDismissRequest = { isDropdownExpanded = false }
                    ) {
                        DropdownMenuItem(
                            onClick = {
                                mapViewModel.clearSelectedGeoElement()
                                mapViewModel.viewModelScope.launch {
                                    mapViewModel.findPlaces(category = MapViewModel.Category.CoffeeShop)
                                    isDropdownExpanded = false
                                }
                            },
                            text = { Text(MapViewModel.Category.CoffeeShop.label) }
                        )
                        DropdownMenuItem(
                            onClick = {
                                mapViewModel.clearSelectedGeoElement()
                                mapViewModel.viewModelScope.launch {
                                    mapViewModel.findPlaces(category = MapViewModel.Category.GasStation)
                                    isDropdownExpanded = false
                                }
                            },
                            text = { Text(MapViewModel.Category.GasStation.label) }
                        )
                        DropdownMenuItem(
                            onClick = {
                                mapViewModel.clearSelectedGeoElement()
                                mapViewModel.viewModelScope.launch {
                                    mapViewModel.findPlaces(category = MapViewModel.Category.Food)
                                    isDropdownExpanded = false
                                }
                            },
                            text = { Text(MapViewModel.Category.Food.label) }
                        )
                        DropdownMenuItem(
                            onClick = {
                                mapViewModel.clearSelectedGeoElement()
                                mapViewModel.viewModelScope.launch {
                                    mapViewModel.findPlaces(category = MapViewModel.Category.Hotel)
                                    isDropdownExpanded = false
                                }
                            },
                            text = { Text(MapViewModel.Category.Hotel.label) }
                        )
                        DropdownMenuItem(
                            onClick = {
                                mapViewModel.clearSelectedGeoElement()
                                mapViewModel.viewModelScope.launch {
                                    mapViewModel.findPlaces(category = MapViewModel.Category.ParksOutdoors)
                                    isDropdownExpanded = false
                                }
                            },
                            text = { Text(MapViewModel.Category.ParksOutdoors.label) }
                        )
                    }

                }

            ) }
    ) {

        MapView(modifier = Modifier
            .fillMaxSize()
            .padding(it),
            arcGISMap = mapViewModel.map,

            onVisibleAreaChanged = { newVisibleArea ->
                mapViewModel.geoViewExtent = newVisibleArea.extent
            },
            mapViewProxy = mapViewModel.mapViewProxy,
            graphicsOverlays = listOf(mapViewModel.graphicsOverlay),
            onSingleTapConfirmed = { singleTapConfirmedEvent ->
                mapViewModel.identify(singleTapConfirmedEvent)
            },

            content = if (selectedGeoElement != null) {

                {
                    Callout(
                        modifier = Modifier
                            .wrapContentSize()
                            .height(120.dp)
                            .widthIn(max = 300.dp),
                        geoElement = selectedGeoElement,
                        tapLocation = mapViewModel.tapLocation.value
                    ) {
                        CalloutContent(
                            selectedElementAttributes = selectedGeoElement.attributes
                        )
                    }
                }
            } else {
                null
            }

        )

    }

}

/**
 * Content for the Callout to display information on the tapped graphics overlay and its [selectedElementAttributes]
 */
@Composable
fun CalloutContent(
    selectedElementAttributes: Map<String, Any?>
) {
    LazyColumn(contentPadding = PaddingValues(8.dp)) {
        selectedElementAttributes.forEach { attribute ->
            item {
                val style = if (attribute.key == "PlaceName") {
                    MaterialTheme.typography.titleLarge
                } else {
                    MaterialTheme.typography.bodyMedium
                }

                Text(
                    text = "${attribute.value}",
                    fontStyle = FontStyle.Normal,
                    style = style,
                    textAlign = TextAlign.Start
                )

            }
        }
    }
}





